/***
*		branch.c -- configurable branching tool for b&p
*		 1997 Michael Rees (aka Another Option)   All rights reserved.
***/

/*
=============================================================================
|   DATE   | WHAT
=============================================================================
  1 Sep 97 | created
  2 Sep 97 | added list processing; check tool->track to see if we're in a macro
           | that's being created; truncate menu when sysex is not valid
  3 Sep 97 | icon change; init to DUPE, dope, not PASS
 14 Sep 97 | added branch array; can't change destinations in a sysex unless you
           | keep track of what branched, dipshit.
 15 Nov 97 | foreign sysex was not being passed to the next tool.
 14 Jan 98 | unbranched noteoffs were being munged in processevent() -- not
           | enough curly braces... fixed? was it really broken??
  3 Mar 98 | foreign sysex not passed if USE_SYSEX was set
*/


#define DBUG 0
//#define FILEDUMP

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <intuition/intuition.h>
#include <string.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
#include <stdarg.h>

#include "bars.h"
#include "BarsMacros.h"
#include "BPLib.h"
//#include "configlib.h"
#include "proto/branch_protos.h"
#include "debug.h"
#include "branchw.h"
#include "/icons/branchdown.h" // these are the regular icons
#include "/icons/branchup.h"
//#include "/icons/megabranch.h" // this has the drop shadow icons
#include "snapshotgadget.h"
#include "ToolID.h"
#define TOOLID MBra_TOOLID

static UBYTE const versiontag[] = "$VER: MegaBranch 40.14 (03.03.98)";

#define SX_LENGTH (TOOL_SIZE-sizeof(struct Tool))

struct BranchTool {
	struct Tool tool;
	UBYTE flags;  // sysex flags &c (eventually no doubt)
	UBYTE notes;  // flags for what do do with events
	UBYTE bend;
	UBYTE mtouch;
	UBYTE ptouch;
	UBYTE pgmchg;
	UBYTE ccnum;  // currently-displayed cc num
	UBYTE cc[121];
	BYTE branch[128]; // what notes got branched
};

/* flag bits */
#define FREE_SYSEX 1
#define USE_SYSEX  2

/* flags for what to do with events */
/* NOTE that these correspond with the menuitem numbers so don't screw it up! */
#define PASS   0
#define BRANCH 1
#define DUPE   2

#define TOOL_SIZE (sizeof(struct BranchTool) - 128) // don't save branch array

extern struct IntuitionBase *IntuitionBase;
extern struct GfxBase *GfxBase;
extern struct LayersBase *LayersBase;
extern struct DOSBase *DOSBase;
extern struct ExecBase *SysBase;
extern struct Functions *functions;

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void __regargs inittool(struct BranchTool *tool)

{
	memset((char *)&tool->flags,DUPE,TOOL_SIZE - sizeof(struct Tool));
	memset(tool->branch,0,128);
	tool->ccnum = 65;
	tool->flags = FREE_SYSEX;
	tool->tool.touched |= TOUCH_INIT;
}


#define SECRET_MESSAGE 9999 // mn_Length of message to edittool() for refresh

/*==========================================================================
* dosysex
*
*  INPUT: pointer to StringEvent
* RETURN: pointer to Event
*
* EFFECT: gets new tool state from sysex message; signal ui to refresh via
*         a secret message
*=========================================================================*/

static __inline struct Event *dosysex(struct BranchTool *tool,
                                      struct StringEvent *event)

{
	register unsigned char *sxdata;
	struct Message *msg;
	UWORD sxflag;
	UBYTE ccnum; // remember current ccnum instead of changing to sysex value

	event->tool = tool->tool.next;
	if (tool->flags & USE_SYSEX) {
		if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
			sxflag = tool->flags & (USE_SYSEX | FREE_SYSEX);
			ccnum = tool->ccnum;
			memcpy((void *)&tool->flags,(void *)sxdata,SX_LENGTH);
			if (sxflag & USE_SYSEX) tool->flags |= USE_SYSEX;
			if (sxflag & FREE_SYSEX) tool->flags |= FREE_SYSEX;
			tool->ccnum = ccnum;
			if (!InToolPad(&tool->tool)) {
				if (tool->tool.window) {
					if (msg=(struct Message *)functions->myalloc(sizeof(struct Message),0)) {
						msg->mn_Node.ln_Type = NT_MESSAGE;
						msg->mn_ReplyPort = NULL;
						msg->mn_Length = SECRET_MESSAGE;
						PutMsg(tool->tool.window->UserPort,msg);
					}
				}
				if (tool->flags & FREE_SYSEX) {
					functions->freestring(event->string);
					functions->freeevent((struct Event *)event);
					event=NULL;
				}
			}
		}
	}
	return((struct Event *)event);
}

/* event->type signal to free this event */
#define EVENT_EATME 0xFF

/*===========================================================================
* route the event to an appropriate destination, cloning it if necessary, and
* tracking whether a note got branched. depends on caller to pass events in
* deference to processlist()
*============================================================================*/

static struct Event * __regargs doit(struct BranchTool *tool,
                                     struct Event *event, UBYTE flag)
{
	if (flag == BRANCH) {
		if (tool->tool.branch) {
			event->tool = tool->tool.branch;
			event->type |= EVENT_BRANCH;
			if (IsNoteOn(event)) tool->branch[event->byte1]++;
		}
		else {
			event->type = EVENT_EATME;
		}
	}
	else /* if (flag == DUPE) */ {
		if (tool->tool.branch) {
			event->next = functions->dupeevent(event);
			if (event->next) {
				event->next->next = NULL;
				event->next->tool = tool->tool.branch;
				event->next->type |= EVENT_BRANCH;
				if (IsNoteOn(event)) tool->branch[event->byte1]++;
			}
		}
		event->tool = tool->tool.next; // pass if no branch connected
	}
	return(event);
}
/*==========================================================================
* processevent
*
* may send extraneous noteoffs if note routing changes, but otherwise
* it's fairly efficient.
*=========================================================================*/

static struct Event * __stdargs __saveds processevent(struct Event *event)

{
   struct BranchTool *tool = (struct BranchTool *)event->tool;

	if (!tool->tool.touched) inittool(tool);
	if (IsNoteOff(event)) {
		if (tool->branch[event->byte1]) {
			if (tool->tool.branch) {
				if (event->next = (struct Event *)CloneNoteOff((struct NoteEvent *)event)) {
					event->next->tool = event->tool->branch;
					event->tool = event->tool->next;
					event->next->type |= EVENT_BRANCH;
					tool->branch[event->byte1]--;
				}
			}
			else {
				tool->branch[event->byte1]--;
				event->tool = event->tool->next;
			}
		}
		else event->tool = event->tool->next;
	}
	else if (IsAllOff(event)) {
		if (tool->tool.branch) {
			event->next = functions->dupeevent(event);
			if (event->next) {
				event->next->type |= EVENT_BRANCH;
				event->next->tool = event->tool->branch;
			}
			event->tool = event->tool->next;
		}
		memset((char *)tool->branch,0,128);
	}
	else if (IsNoteOn(event) && tool->notes) {
		event = doit(tool,event,tool->notes);
	}
	else if (IsPitchBend(event) && tool->bend) event = doit(tool,event,tool->bend);
	else if (IsMonoTouch(event) && tool->mtouch) event = doit(tool,event,tool->mtouch);
	else if (IsPolyTouch(event) && tool->ptouch) event = doit(tool,event,tool->ptouch);
	else if (IsPgmChg(event) && tool->pgmchg) event = doit(tool,event,tool->pgmchg);
	else if (IsCtrlChg(event) && tool->cc[event->byte1]) 
		event = doit(tool,event,tool->cc[event->byte1]);
	else if (IsSysEx(event)) event = dosysex(tool,(struct StringEvent *)event);
	else event->tool = event->tool->next;
	if (event->type == EVENT_EATME) {
		functions->freeevent(event);
		event = NULL;
	}
	return(event);
}

/*==========================================================================
* processlist
*=========================================================================*/

static struct Event * __stdargs __saveds processlist(struct BranchTool *tool,
                                                     struct Event *list,
                                                     long start, long stop)

{
	struct Event *event = list;

	if (!tool->tool.touched) inittool(tool);
	if (start < 0) start = 0;
	while (event && (event->time < start)) event = event->next;
	for ( ; event && (event->time <= stop) ; event = event->next ) {
		if (IsNote(event) && tool->notes) event = doit(tool,event,tool->notes);
		else if (IsPitchBend(event) && tool->bend) event = doit(tool,event,tool->bend);
		else if (IsMonoTouch(event) && tool->mtouch) event = doit(tool,event,tool->mtouch);
		else if (IsPolyTouch(event) && tool->ptouch) event = doit(tool,event,tool->ptouch);
		else if (IsPgmChg(event) && tool->pgmchg) event = doit(tool,event,tool->pgmchg);
		else if (IsCtrlChg(event) && tool->cc[event->byte1]) 
			event = doit(tool,event,tool->cc[event->byte1]);
/* this doesn't work for some fucking reason. who needs it anyway? */
/* ps: sysex events don't get a PADEDIT bit */
//		else if (IsSysEx(event)) event = dosysex(tool,(struct StringEvent *)event);
		else event->tool = tool->tool.next;
		if (event->type == EVENT_EATME) event->time = -1;
	}
	event = list = functions->sorteventlist(list);
	for ( ; event ; event=list ) {
		if (event->time >= 0) break;
		/* else */
		list = event->next;
		functions->freeevent(event);
	}
	return(list);
}

/*==========================================================================
* dosnapshot
*=========================================================================*/

static void __regargs dosnapshot(struct BranchTool *tool)

{
	UBYTE sx[SX_LENGTH];
	struct StringEvent *event;
	long time = functions->running ?
	                         (functions->timenow ? functions->timenow-1 : 0) : -1;

	memcpy(sx,&tool->flags,SX_LENGTH);
	if (event = BuildSnapshot(&tool->tool, time, sx, SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
	}
}

/*==========================================================================
* call popupmenu and stash a valid return in the flag 
* THIS is why it's important to keep menuitem numbes & routing flags in sync!!
*=========================================================================*/

static UBYTE __regargs dopop(struct Window *w,UBYTE *flag)
{
	short menunum = (short)functions->PopUpMenu(w,&PopPipeMenu);

	menunum = ITEMNUM(menunum);
	if (menunum < 3) {
		*flag = menunum;
		return(1);
	}
	else return(0);
}

struct BranchTool *selectedtool = NULL;

/*==========================================================================
* update the cc popup gadget image while scrolling the ccnum gadget *
*=========================================================================*/
static long __stdargs __saveds hiloPix(struct Window *w,struct Gadget *g,long h)

{
   char hiloText[4];
	struct Gadget *popgad;

   SetRPort(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset,
            HILO_APEN, HILO_BPEN,HILO_DRAWMODE);
#pragma msg 104 ignore push
	sprintf(hiloText,"%3ld",h);
#pragma msg 104 ignore pop
	Text(w->RPort,hiloText,3);
	if (popgad = functions->GetGadget(w,CCPOP)) {
		if (selectedtool->cc[h] == BRANCH) popgad->GadgetRender = (APTR)&ElbowImage;
		else if (selectedtool->cc[h] == DUPE) popgad->GadgetRender = (APTR)&BranchImage;
		else popgad->GadgetRender = (APTR)&PipeImage;
		functions->RefreshGadget(w,CCPOP);
	}
	return(h);
}

#define tw tool->tool.window
#define REFRESH_CC 2

/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __stdargs __saveds edittool(struct BranchTool *tool)

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	struct Menu *menu;
	struct MenuItem *menuitem;
//	short menunum;
	UBYTE refresh;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);

	/* if no tool.track, we are in a macrotool */
	if ((refresh = InToolPad(&tool->tool)) || !tool->tool.track) refresh = 0;
	else refresh = 20;
	nw=OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,refresh);
	if (!nw) return;
	if (refresh) {
		AddSnapshotGadget(tw);
		if (tool->flags & USE_SYSEX) UseSysexItem.Flags |= CHECKED;
		else UseSysexItem.Flags &= ~CHECKED;
		if (tool->flags & FREE_SYSEX) FreeSysexItem.Flags |= CHECKED;
		else FreeSysexItem.Flags &= ~CHECKED;
		MenuItem21.NextItem = &MenuItem22;
	}
	else {
		MenuItem21.NextItem = NULL; //no sysex stuff in the menu in toolpad
	}
	if (menu = functions->DupeMenu(&MenuList1)) SetMenuStrip(tw,menu);
	else SetWindowTitles(tw,(UBYTE *)-1,MenuList1.MenuName);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	functions->FatEmbossedPropOn(tw,CCNUM,CCNUM+GDOWN,CCNUM+GUP,hiloPix,121,1);
	refresh = 255;

	/*** ui handling ***/
	for ( ; ; ) {
		if (refresh) {
			if (refresh & REFRESH_CC) {
				selectedtool = tool;
				functions->ModifyEmbossedProp(tw,CCNUM,tool->ccnum,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,CCNUM);
				gadget = functions->GetGadget(tw,CCPOP);
				if (tool->cc[tool->ccnum] == BRANCH) gadget->GadgetRender = (APTR)&ElbowImage;
				else if (tool->cc[tool->ccnum] == DUPE) gadget->GadgetRender = (APTR)&BranchImage;
				else gadget->GadgetRender = (APTR)&PipeImage;
				functions->RefreshGadget(tw,CCPOP);
				refresh &= ~REFRESH_CC;
			}

			if (refresh) {
				gadget = functions->GetGadget(tw,PGMPOP);
				if (tool->pgmchg == BRANCH) gadget->GadgetRender = (APTR)&ElbowImage;
				else if (tool->pgmchg == DUPE) gadget->GadgetRender = (APTR)&BranchImage;
				else gadget->GadgetRender = (APTR)&PipeImage;
				functions->RefreshGadget(tw,PGMPOP);

				gadget = functions->GetGadget(tw,PTOUCHPOP);
				if (tool->ptouch == BRANCH) gadget->GadgetRender = (APTR)&ElbowImage;
				else if (tool->ptouch == DUPE) gadget->GadgetRender = (APTR)&BranchImage;
				else gadget->GadgetRender = (APTR)&PipeImage;
				functions->RefreshGadget(tw,PTOUCHPOP);

				gadget = functions->GetGadget(tw,MTOUCHPOP);
				if (tool->mtouch == BRANCH) gadget->GadgetRender = (APTR)&ElbowImage;
				else if (tool->mtouch == DUPE) gadget->GadgetRender = (APTR)&BranchImage;
				else gadget->GadgetRender = (APTR)&PipeImage;
				functions->RefreshGadget(tw,MTOUCHPOP);

				gadget = functions->GetGadget(tw,BENDPOP);
				if (tool->bend == BRANCH) gadget->GadgetRender = (APTR)&ElbowImage;
				else if (tool->bend == DUPE) gadget->GadgetRender = (APTR)&BranchImage;
				else gadget->GadgetRender = (APTR)&PipeImage;
				functions->RefreshGadget(tw,BENDPOP);

				gadget = functions->GetGadget(tw,NOTEPOP);
				if (tool->notes == BRANCH) gadget->GadgetRender = (APTR)&ElbowImage;
				else if (tool->notes == DUPE) gadget->GadgetRender = (APTR)&BranchImage;
				else gadget->GadgetRender = (APTR)&PipeImage;
				functions->RefreshGadget(tw,NOTEPOP);
				refresh = 0;
			}
		}
		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		if (message->ExecMessage.mn_Length == SECRET_MESSAGE) { // ooh, read sysex!
			refresh = 0xFF;
			functions->myfree((char *)message,sizeof(struct Message));
			continue;
		}
		/* else */
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) {
			selectedtool = NULL;
			break;
		}

		else if (class == MENUPICK) {
			while (code != MENUNULL) {
				menuitem = ItemAddress(menu,code);
				switch (ITEMNUM(code)) {
					case 0 : // elbow all ccs
						memset((char *)&tool->cc[0],BRANCH,121);
						refresh = REFRESH_CC;
						break;
					case 1 : // pass all ccs
						memset((char *)&tool->cc[0],PASS,121);
						refresh = REFRESH_CC;
						break;
					case 2 : // branch all ccs
						memset((char *)&tool->cc[0],DUPE,121);
						refresh = REFRESH_CC;
						break;
					case 3 : // use sysex
						tool->flags ^= USE_SYSEX;
						break;
					case 4 : // free sysex
						tool->flags ^= FREE_SYSEX;
						break;
					case 5 : // snap
						dosnapshot(tool);
						break;
					default : break;
				}
				code = menuitem->NextSelect;
			}
		}

		else if (class == GADGETUP) {
			if (gadget->GadgetID == SNAPSHOT_GADGET_ID) {
				dosnapshot(tool);
			}
		}

		else if (class == GADGETDOWN) {
			selectedtool = tool;
			switch (gadget->GadgetID) {
				case CCNUM :
					tool->ccnum = functions->DragEmbossedProp(tw,CCNUM);
					refresh = REFRESH_CC;
					break;
				case CCNUM+GUP :
					tool->ccnum = functions->ShiftEmbossedProp(tw,CCNUM,1,0);
					refresh = REFRESH_CC;
					break;
				case CCNUM+GDOWN :
					tool->ccnum = functions->ShiftEmbossedProp(tw,CCNUM,-1,0);
					refresh = REFRESH_CC;
					break;
				case NOTEPOP :
					refresh = dopop(tw,&tool->notes);
					break;
				case BENDPOP :
					refresh = dopop(tw,&tool->bend);
					break;
				case MTOUCHPOP :
					refresh = dopop(tw,&tool->mtouch);
					break;
				case PTOUCHPOP :
					refresh = dopop(tw,&tool->ptouch);
					break;
				case PGMPOP :
					refresh = dopop(tw,&tool->pgmchg);
					break;
				case CCPOP :
					refresh = dopop(tw,&tool->cc[tool->ccnum]);
					if (refresh) refresh = REFRESH_CC;
					break;
			}
		}
	}

	functions->FatEmbossedPropOff(tw,CCNUM,CCNUM+GDOWN,CCNUM+GUP);
	NewCloseEmbossedWindow(&tool->tool,nw);
	if (menu) functions->DeleteMenu(menu);
}

/*****************************************/
static struct Tool * __stdargs __saveds createtool(void)
{
	struct BranchTool *tool;
	
	tool = (struct BranchTool *)functions->myalloc(sizeof(struct BranchTool),MEMF_CLEAR);
//	if (tool) memset((char *)tool,0,sizeof(struct BranchTool));
	return ((struct Tool *)tool);
}

/*****************************************/
static void __stdargs __saveds deletetool(struct BranchTool *tool)
{
	functions->myfree((char *)tool,sizeof(struct BranchTool));
}

/*****************************************/
static struct Tool * __stdargs __saveds loadtool(long file, long size)
{
	struct BranchTool *tool;

	tool = (struct BranchTool *)functions->myalloc(sizeof(struct BranchTool),MEMF_CLEAR);
	if (tool) {
		functions->fastread(file,(char *)tool,size);
		size -= TOOL_SIZE;
		memset((char *)tool->branch,0,128);
	}
	if (size > 0) functions->fastseek(file,size,0);
	return((struct Tool *)tool);
}

struct ToolMaster master;

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to ToolMaster
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __stdargs __saveds inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = TOOLID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.processlist = processlist;
	master.createtool = createtool;
	master.deletetool = deletetool;
//	master.savetool = savetool;
	master.loadtool = loadtool;
	master.image = &ImageDown;
	master.upimage = &ImageUp;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL | TOOL_BRANCHOUT;
	return(&master);
}
